@@ -1,4 +1,4 @@
-# Copyright (C) 2003 Nominum, Inc.
+# Copyright (C) 2003-2005 Nominum, Inc.
 #
 # Permission to use, copy, modify, and distribute this software and its
 # documentation for any purpose with or without fee is hereby granted,
@@ -13,8 +13,6 @@
 # ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 # OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
-# $Id: zone.py,v 1.34 2003/08/02 21:42:39 halley Exp $
-
 """DNS Zones."""
 
 from __future__ import generators
@@ -28,7 +26,20 @@ import dns.rdataclass
 import dns.rdatatype
 import dns.rdata
 import dns.tokenizer
-        
+import dns.ttl
+
+class BadZone(dns.exception.DNSException):
+    """The zone is malformed."""
+    pass
+
+class NoSOA(BadZone):
+    """The zone has no SOA RR at its origin."""
+    pass
+
+class NoNS(BadZone):
+    """The zone has no NS RRset at its origin."""
+    pass
+
 class Zone(object):
     """A DNS zone.
 
@@ -46,15 +57,17 @@ class Zone(object):
     @ivar nodes: A dictionary mapping the names of nodes in the zone to the
     nodes themselves.
     @type nodes: dict
+    @ivar relativize: should names in the zone be relativized?
+    @type relativize: bool
     @cvar node_factory: the factory used to create a new node
     @type node_factory: class or callable
     """
 
     node_factory = dns.node.Node
 
-    __slots__ = ['rdclass', 'origin', 'nodes']
+    __slots__ = ['rdclass', 'origin', 'nodes', 'relativize']
     
-    def __init__(self, origin, rdclass=dns.rdataclass.IN):
+    def __init__(self, origin, rdclass=dns.rdataclass.IN, relativize=True):
         """Initialize a zone object.
 
         @param origin: The origin of the zone.
@@ -65,6 +78,7 @@ class Zone(object):
         self.rdclass = rdclass
         self.origin = origin
         self.nodes = {}
+        self.relativize = relativize
 
     def __eq__(self, other):
         """Two zones are equal if they have the same origin, class, and
@@ -93,9 +107,12 @@ class Zone(object):
         elif not isinstance(name, dns.name.Name):
             raise KeyError, \
                   "name parameter must be convertable to a DNS name"
-        if name.is_absolute() and not name.is_subdomain(self.origin):
-            raise KeyError, \
-                  "name parameter must be a subdomain of the zone origin"
+        if name.is_absolute():
+            if not name.is_subdomain(self.origin):
+                raise KeyError, \
+                      "name parameter must be a subdomain of the zone origin"
+            if self.relativize:
+                name = name.relativize(self.origin)
         return name
     
     def __getitem__(self, key):
@@ -221,7 +238,7 @@ class Zone(object):
             rdtype = dns.rdatatype.from_text(rdtype)
         if isinstance(covers, str):
             covers = dns.rdatatype.from_text(covers)
-        node = self.get_node(name, create)
+        node = self.find_node(name, create)
         return node.find_rdataset(self.rdclass, rdtype, covers, create)
 
     def get_rdataset(self, name, rdtype, covers=dns.rdatatype.NONE,
@@ -481,6 +498,22 @@ class Zone(object):
             if want_close:
                 f.close()
 
+    def check_origin(self):
+        """Do some simple checking of the zone's origin.
+
+        @raises dns.zone.NoSOA: there is no SOA RR
+        @raises dns.zone.NoNS: there is no NS RRset
+        @raises KeyError: there is no origin node
+        """
+        if self.relativize:
+            name = dns.name.empty
+        else:
+            name = self.origin
+        if self.get_rdataset(name, dns.rdatatype.SOA) is None:
+            raise NoSOA
+        if self.get_rdataset(name, dns.rdatatype.NS) is None:
+            raise NoNS
+
 
 class _MasterReader(object):
     """Read a DNS master file
@@ -497,9 +530,17 @@ class _MasterReader(object):
     @type relativize: bool
     @ivar zone: the zone
     @type zone: dns.zone.Zone object
+    @ivar saved_state: saved reader state (used when processing $INCLUDE)
+    @type saved_state: list of (tokenizer, current_origin, last_name, file)
+    tuples.
+    @ivar current_file: the file object of the $INCLUDed file being parsed
+    (None if no $INCLUDE is active).
+    @ivar allow_include: is $INCLUDE allowed?
+    @type allow_include: bool
     """
 
-    def __init__(self, tok, origin, rdclass, relativize, zone_factory=Zone):
+    def __init__(self, tok, origin, rdclass, relativize, zone_factory=Zone,
+                 allow_include=False):
         if isinstance(origin, str):
             origin = dns.name.from_text(origin)
         self.tok = tok
@@ -507,7 +548,10 @@ class _MasterReader(object):
         self.relativize = relativize
         self.ttl = 0
         self.last_name = None
-        self.zone = zone_factory(origin, rdclass)
+        self.zone = zone_factory(origin, rdclass, relativize=relativize)
+        self.saved_state = []
+        self.current_file = None
+        self.allow_include = allow_include
 
     def _eat_line(self):
         while 1:
@@ -521,6 +565,13 @@ class _MasterReader(object):
         token = self.tok.get(want_leading = True)
         if token[0] != dns.tokenizer.WHITESPACE:
             self.last_name = dns.name.from_text(token[1], self.current_origin)
+        else:
+            token = self.tok.get()
+            if token[0] == dns.tokenizer.EOL or \
+               token[0] == dns.tokenizer.EOF:
+                # treat leading WS followed by EOL/EOF as if they were EOL/EOF.
+                return
+            self.tok.unget(token)
         name = self.last_name
         if not name.is_subdomain(self.zone.origin):
             self._eat_line()
@@ -532,13 +583,11 @@ class _MasterReader(object):
             raise dns.exception.SyntaxError
         # TTL
         try:
-            ttl = int(token[1], 0)
+            ttl = dns.ttl.from_text(token[1])
             token = self.tok.get()
             if token[0] != dns.tokenizer.IDENTIFIER:
                 raise dns.exception.SyntaxError
-        except dns.exception.SyntaxError:
-            raise dns.exception.SyntaxError
-        except:
+        except dns.ttl.BadTTL:
             ttl = self.ttl
         # Class
         try:
@@ -553,46 +602,119 @@ class _MasterReader(object):
         if rdclass != self.zone.rdclass:
             raise dns.exception.SyntaxError, "RR class is not zone's class"
         # Type
-        rdtype = dns.rdatatype.from_text(token[1])
+        try:
+            rdtype = dns.rdatatype.from_text(token[1])
+        except:
+            raise dns.exception.SyntaxError, \
+                  "unknown rdatatype '%s'" % token[1]
         n = self.zone.nodes.get(name)
         if n is None:
             n = self.zone.node_factory()
             self.zone.nodes[name] = n
-        rd = dns.rdata.from_text(rdclass, rdtype, self.tok,
-                                 self.current_origin, False)
+        try:
+            rd = dns.rdata.from_text(rdclass, rdtype, self.tok,
+                                     self.current_origin, False)
+        except dns.exception.SyntaxError:
+            # Catch and reraise.
+            (ty, va) = sys.exc_info()[:2]
+            raise ty, va
+        except:
+            # All exceptions that occur in the processing of rdata
+            # are treated as syntax errors.  This is not strictly
+            # correct, but it is correct almost all of the time.
+            # We convert them to syntax errors so that we can emit
+            # helpful filename:line info.
+            
+            (ty, va) = sys.exc_info()[:2]
+            raise dns.exception.SyntaxError, \
+                  "caught exception %s: %s" % (str(ty), str(va))
+
         rd.choose_relativity(self.zone.origin, self.relativize)
         covers = rd.covers()
         rds = n.find_rdataset(rdclass, rdtype, covers, True)
         rds.add(rd, ttl)
 
     def read(self):
-        """Read a DNS master file and build a zone object."""
+        """Read a DNS master file and build a zone object.
+
+        @raises dns.zone.NoSOA: No SOA RR was found at the zone origin
+        @raises dns.zone.NoNS: No NS RRset was found at the zone origin
+        """
+
+        try:
+            while 1:
+                token = self.tok.get(True, True)
+                if token[0] == dns.tokenizer.EOF:
+                    if not self.current_file is None:
+                        self.current_file.close()
+                    if len(self.saved_state) > 0:
+                        (self.tok,
+                         self.current_origin,
+                         self.last_name,
+                         self.current_file,
+                         self.ttl) = self.saved_state.pop(-1)
+                        continue
+                    break
+                elif token[0] == dns.tokenizer.EOL:
+                    continue
+                elif token[0] == dns.tokenizer.COMMENT:
+                    self.tok.get_eol()
+                    continue
+                elif token[1][0] == '$':
+                    u = token[1].upper()
+                    if u == '$TTL':
+                        token = self.tok.get()
+                        if token[0] != dns.tokenizer.IDENTIFIER:
+                            raise dns.exception.SyntaxError, "bad $TTL"
+                        self.ttl = dns.ttl.from_text(token[1])
+                        self.tok.get_eol()
+                    elif u == '$ORIGIN':
+                        self.current_origin = self.tok.get_name()
+                        self.tok.get_eol()
+                    elif u == '$INCLUDE' and self.allow_include:
+                        token = self.tok.get()
+                        if token[0] != dns.tokenizer.QUOTED_STRING:
+                            raise dns.exception.SyntaxError, \
+                                  "bad filename in $INCLUDE"
+                        filename = token[1]
+                        token = self.tok.get()
+                        if token[0] == dns.tokenizer.IDENTIFIER:
+                            new_origin = dns.name.from_text(token[1], \
+                                                        self.current_origin)
+                            self.tok.get_eol()
+                        elif token[0] != dns.tokenizer.EOL and \
+                             token[0] != dns.tokenizer.EOF:
+                            raise dns.exception.SyntaxError, \
+                                  "bad origin in $INCLUDE"
+                        else:
+                            new_origin = self.current_origin
+                        self.saved_state.append((self.tok,
+                                                 self.current_origin,
+                                                 self.last_name,
+                                                 self.current_file,
+                                                 self.ttl))
+                        self.current_file = file(filename, 'r')
+                        self.tok = dns.tokenizer.Tokenizer(self.current_file,
+                                                           filename)
+                        self.current_origin = new_origin
+                    else:
+                        raise dns.exception.SyntaxError, \
+                              "Unknown master file directive '" + u + "'"
+                    continue
+                self.tok.unget(token)
+                self._rr_line()
+        except dns.exception.SyntaxError, detail:
+            (filename, line_number) = self.tok.where()
+            if detail is None:
+                detail = "syntax error"
+            raise dns.exception.SyntaxError, \
+                  "%s:%d: %s" % (filename, line_number, detail)
         
-        while 1:
-            token = self.tok.get(True, True)
-            if token[0] == dns.tokenizer.EOF:
-                break
-            elif token[0] == dns.tokenizer.EOL:
-                continue
-            elif token[0] == dns.tokenizer.COMMENT:
-                self.tok.get_eol()
-                continue
-            elif token[1][0] == '$':
-                u = token[1].upper()
-                if u == '$TTL':
-                    self.ttl = self.tok.get_int()
-                    self.tok.get_eol()
-                elif u == '$ORIGIN':
-                    self.current_origin = self.tok.get_name()
-                    self.tok.get_eol()
-                else:
-                    self.eat_line()
-                continue
-            self.tok.unget(token)
-            self._rr_line()
+        # Now that we're done reading, do some basic checking of the zone.
+        self.zone.check_origin()
 
 def from_text(text, origin, rdclass = dns.rdataclass.IN, relativize = True,
-              zone_factory=Zone):
+              zone_factory=Zone, filename=None, allow_include=False):
     """Build a zone object from a master file format string.
 
     @param text: the master file format input
@@ -603,19 +725,32 @@ def from_text(text, origin, rdclass = dns.rdataclass.IN, relativize = True,
     @type rdclass: int
     @param relativize: should names be relativized?  The default is True
     @type relativize: bool
-    @rtype: dns.zone.Zone object"""
+    @param zone_factory: The zone factory to use
+    @type zone_factory: function returning a Zone
+    @param filename: The filename to emit when describing where an error
+    occurred; the default is '<string>'.
+    @param allow_include: is $INCLUDE allowed?
+    @type allow_include: bool
+    @type filename: string
+    @raises dns.zone.NoSOA: No SOA RR was found at the zone origin
+    @raises dns.zone.NoNS: No NS RRset was found at the zone origin
+    @rtype: dns.zone.Zone object
+    """
 
     # 'text' can also be a file, but we don't publish that fact
     # since it's an implementation detail.  The official file
     # interface is from_file().
 
-    tok = dns.tokenizer.Tokenizer(text)
-    reader = _MasterReader(tok, origin, rdclass, relativize, zone_factory)
+    if filename is None:
+        filename = '<string>'
+    tok = dns.tokenizer.Tokenizer(text, filename)
+    reader = _MasterReader(tok, origin, rdclass, relativize, zone_factory,
+                           allow_include=allow_include)
     reader.read()
     return reader.zone
 
 def from_file(f, origin, rdclass = dns.rdataclass.IN, relativize = True,
-              zone_factory=Zone):
+              zone_factory=Zone, filename=None, allow_include=True):
     """Read a master file and build a zone object.
 
     @param f: file or string.  If I{f} is a string, it is treated
@@ -626,7 +761,18 @@ def from_file(f, origin, rdclass = dns.rdataclass.IN, relativize = True,
     @type rdclass: int
     @param relativize: should names be relativized?  The default is True
     @type relativize: bool
-    @rtype: dns.zone.Zone object"""
+    @param zone_factory: The zone factory to use
+    @type zone_factory: function returning a Zone
+    @param filename: The filename to emit when describing where an error
+    occurred; the default is '<file>', or the value of I{f} if I{f} is a
+    string.
+    @type filename: string
+    @param allow_include: is $INCLUDE allowed?
+    @type allow_include: bool
+    @raises dns.zone.NoSOA: No SOA RR was found at the zone origin
+    @raises dns.zone.NoNS: No NS RRset was found at the zone origin
+    @rtype: dns.zone.Zone object
+    """
 
     if sys.hexversion >= 0x02030000:
         # allow Unicode filenames; turn on universal newline support
@@ -636,31 +782,44 @@ def from_file(f, origin, rdclass = dns.rdataclass.IN, relativize = True,
         str_type = str
         opts = 'r'
     if isinstance(f, str_type):
+        if filename is None:
+            filename = f
         f = file(f, opts)
         want_close = True
     else:
+        if filename is None:
+            filename = '<file>'
         want_close = False
         
     try:
-        z = from_text(f, origin, rdclass, relativize)
+        z = from_text(f, origin, rdclass, relativize, zone_factory,
+                      filename, allow_include)
     finally:
         if want_close:
             f.close()
     return z
 
-def from_xfr(xfr, zone_factory=Zone):
+def from_xfr(xfr, zone_factory=Zone, relativize=True):
     """Convert the output of a zone transfer generator into a zone object.
     
     @param xfr: The xfr generator
     @type xfr: generator of dns.message.Message objects
-    @rtype: dns.zone.Zone object"""
+    @param relativize: should names be relativized?  The default is True
+    @type relativize: bool
+    @raises dns.zone.NoSOA: No SOA RR was found at the zone origin
+    @raises dns.zone.NoNS: No NS RRset was found at the zone origin
+    @rtype: dns.zone.Zone object
+    """
     
     z = None
     for r in xfr:
         if z is None:
-            origin = r.answer[0].name
+            if relativize:
+                origin = r.origin
+            else:
+                origin = r.answer[0].name
             rdclass = r.answer[0].rdclass
-            z = zone_factory(origin, rdclass)
+            z = zone_factory(origin, rdclass, relativize=relativize)
         for rrset in r.answer:
             znode = z.nodes.get(rrset.name)
             if not znode:
@@ -668,5 +827,9 @@ def from_xfr(xfr, zone_factory=Zone):
                 z.nodes[rrset.name] = znode
             zrds = znode.find_rdataset(rrset.rdclass, rrset.rdtype,
                                        rrset.covers, True)
-            zrds.update(rrset)
+            zrds.update_ttl(rrset.ttl)
+            for rd in rrset:
+                rd.choose_relativity(z.origin, relativize)
+                zrds.add(rd)
+    z.check_origin()
     return z
